# -*- coding: utf-8 -*-
# cython:language_level=3
"""
-------------------------------------------------
   File Name：     setupPylcc
   Description :
   Author :       liaozhaoyan
   date：          2022/1/17
-------------------------------------------------
   Change Activity:
                   2022/1/17:
-------------------------------------------------
"""
__author__ = 'liaozhaoyan'

VERSION = '0.2.5'

from setuptools import setup

setup(name='pylcc',
      version=VERSION,
      description="pylcc is short for python libbpf compile collections",
      long_description='pylcc is short for python libbpf compile collections',
      classifiers=["Topic :: System :: Operating System Kernels :: Linux",
                   "Programming Language :: Python",
                   "Programming Language :: Python :: 2",
                   "Programming Language :: Python :: 2.7",
                   "Programming Language :: Python :: 3",
                   "Programming Language :: Python :: 3.5",
                   "Programming Language :: Python :: 3.6",
                   "Programming Language :: Python :: 3.7",
                   "Programming Language :: Python :: 3.8",
                   "Programming Language :: Python :: 3.9",
                   "Programming Language :: Python :: 3.10",
                   "Programming Language :: Python :: Implementation :: PyPy",
                   ],  # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='linux kernel trace',
      author='liaozhaoyan',
      author_email='zhaoyan.liao@linux.alibaba.com',
      url="https://gitee.com/anolis/surftrace",
      license='MIT',
      packages=["pylcc"],
      include_package_data=True,
      zip_safe=True,
      install_requires=['surftrace>=0.6'],
      )

if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
# cython:language_level=3
"""
-------------------------------------------------
   File Name：     lbcMaps
   Description :
   Author :       liaozhaoyan
   date：          2021/7/20
-------------------------------------------------
   Change Activity:
                   2021/7/20:
-------------------------------------------------
"""
__author__ = 'liaozhaoyan'

import ctypes as ct
import struct
from sys import version_info

try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping
from surftrace import InvalidArgsException


class CtypeData(object):
    def __init__(self, dForm, data):
        super(CtypeData, self).__init__()
        self.dForm = dForm
        self._initData(data)

    def _invData(self, v, dFormat):
        # if dFormat.has_key('format'):  not support for python3
        if 'format' in dFormat:
            if dFormat['array']:
                data = None
                for c in v:
                    data += struct.pack("%s" % dFormat['format'][0], c)
            else:
                data = struct.pack("%s" % dFormat['format'], v)
        else:
            data = None
            for d in dFormat['cells']:
                data += self._invData(v[d['member']], d)
        return data

    @staticmethod
    def __strlen(stream, size):
        for i in range(size):
            if stream[i] == '\0':
                return i
        return i

    def _loadData(self, data, start, dFormat):
        if "format" in dFormat:
            if dFormat['array'] > 0:
                if dFormat['format'][0] == 'c':
                    stream = data[start:start + dFormat['size'] * dFormat['array']]
                    size = self.__strlen(stream, dFormat['array'])
                    tp = struct.unpack("%ds" % size, stream[:size])[0]
                    s = ''
                    if version_info.major == 2:
                        s = ''.join(tp)
                    else:
                        for t in tp:
                            if t != 0:
                                s += chr(t)
                            else:
                                break
                    return s
                else:
                    return struct.unpack(dFormat['format'].encode('utf-8'),
                                         data[start:start + dFormat['size'] * dFormat['array']])
            else:
                return struct.unpack(dFormat['format'].encode('utf-8'), data[start:start + dFormat['size']])[0]
        else:
            rDict = {}
            for d in dFormat['cells']:
                beg = start + d['offset']
                rDict[d['member']] = self._loadData(data[beg:], start, d)
            return rDict

    def _initData(self, data):
        if "format" in self.dForm:
            d = {'value': self._loadData(data, 0, self.dForm)}
        else:
            d = self._loadData(data, 0, self.dForm)
        self.localDict = d
        self.__dict__.update(self.localDict)

class CtypeTable(CtypeData):
    def __init__(self, dForm, data=None):
        super(CtypeTable, self).__init__(dForm, data)
        self.localData = []

    def _initData(self, data):
        pass

    def add(self, data):
        l = len(data)
        v = ct.create_string_buffer(l)
        ct.memmove(v, data, l)
        self.localData.append(v)

    def clear(self):
        self.localData = []

    def output(self):
        l = []
        for r in self.localData:
            l.append(self._loadData(r, 0, self.dForm))
        return l

    def input(self, k):
        return self._invData(k, self.dForm)

    def load(self, v):
        return self._loadData(v, 0, self.dForm)


class CeventBase(object):
    def __init__(self, so, name):
        self._so = so
        self._so.lbc_bpf_get_maps_id.restype = ct.c_int
        self._so.lbc_bpf_get_maps_id.argtypes = [ct.c_char_p]
        self._id = self._so.lbc_bpf_get_maps_id(ct.c_char_p(name.encode('utf-8')))
        if self._id < 0:
            raise InvalidArgsException("map %s, not such event" % name)
        self.name = name
        super(CeventBase, self).__init__()


class CtableBase(CeventBase):
    def __init__(self, so, name, dTypes):
        super(CtableBase, self).__init__(so, name)
        self._kd = dTypes['ktype']
        self._vd = dTypes['vtype']
        self.keys = CtypeTable(self._kd)
        self.values = CtypeTable(self._vd)
        self.__setup_localCalls()

    def __setup_localCalls(self):
        self._so.lbc_map_lookup_elem.restype = ct.c_int
        self._so.lbc_map_lookup_elem.argtypes = [ct.c_int, ct.c_void_p, ct.c_void_p]
        self._so.lbc_map_lookup_and_delete_elem.restype = ct.c_int
        self._so.lbc_map_lookup_and_delete_elem.argtypes = [ct.c_int, ct.c_void_p, ct.c_void_p]
        self._so.lbc_map_delete_elem.restype = ct.c_int
        self._so.lbc_map_delete_elem.argtypes = [ct.c_int, ct.c_void_p]
        self._so.lbc_map_get_next_key.restype = ct.c_int
        self._so.lbc_map_get_next_key.argtypes = [ct.c_int, ct.c_void_p, ct.c_void_p]

    def _getSize(self, dCell):
        if dCell['array']:
            return dCell['array'] * dCell['size']
        return dCell['size']

    def get(self):
        ksize = self._getSize(self._kd)
        vsize = self._getSize(self._vd)

        self.keys.clear()
        self.values.clear()
        v = ct.create_string_buffer(vsize)
        k1 = ct.create_string_buffer(ksize)
        k2 = ct.create_string_buffer(ksize)
        while self._so.lbc_map_get_next_key(self._id, k1, k2) == 0:
            self.keys.add(k2)
            self._so.lbc_map_lookup_elem(self._id, k2, v)
            self.values.add(v)
            ct.memmove(k1, k2, ksize)
        return dict(zip(self.keys.output(), self.values.output()))

    def getThenClear(self):
        ksize = self._getSize(self._kd)
        vsize = self._getSize(self._vd)

        self.keys.clear()
        self.values.clear()
        v = ct.create_string_buffer(vsize)
        k1 = ct.create_string_buffer(ksize)
        k2 = ct.create_string_buffer(ksize)
        while self._so.lbc_map_get_next_key(self._id, k1, k2) == 0:
            self.keys.add(k2)
            r = self._so.lbc_map_lookup_and_delete_elem(self._id, k2, v)
            if r < 0:
                raise InvalidArgsException(
                    "lbc_map_lookup_and_delete_elem return %d, os may not support this opertation." % r)
            self.values.add(v)
            ct.memmove(k1, k2, ksize)
        r = dict(zip(self.keys.output(), self.values.output()))
        return r

    def getKeys(self):
        self.keys.clear()
        self.values.clear()
        ksize = self._getSize(self._kd)

        k1 = ct.create_string_buffer(ksize)
        k2 = ct.create_string_buffer(ksize)
        while self._so.lbc_map_get_next_key(self._id, k1, k2) == 0:
            self.keys.add(k2)
            ct.memmove(k1, k2, ksize)
        return self.keys.output()

    def getKeyValue(self, k):
        vsize = self._getSize(self._vd)
        key = self.keys.input(k)
        v = ct.create_string_buffer(vsize)
        if self._so.lbc_map_lookup_elem(self._id, key, v) == 0:
            return self.values.load(v)
        return None

    def clear(self):
        ksize = self._kd['size']
        k1 = ct.create_string_buffer(ksize)
        k2 = ct.create_string_buffer(ksize)
        while self._so.lbc_map_get_next_key(self._id, k1, k2) == 0:
            self._so.lbc_map_delete_elem(self._id, k2)
            ct.memmove(k1, k2, ksize)


class CmapsHash(CtableBase):
    def __init__(self, so, name, dTypes):
        super(CmapsHash, self).__init__(so, name, dTypes)


class CmapsArray(CtableBase):
    def __init__(self, so, name, dTypes):
        super(CmapsArray, self).__init__(so, name, dTypes)


class CmapsLruHash(CtableBase):
    def __init__(self, so, name, dTypes):
        super(CmapsHash, self).__init__(so, name, dTypes)


class CmapsPerHash(CtableBase):
    def __init__(self, so, name, dTypes):
        super(CmapsPerHash, self).__init__(so, name, dTypes)


class CmapsPerArray(CtableBase):
    def __init__(self, so, name, dTypes):
        super(CmapsPerArray, self).__init__(so, name, dTypes)


class CmapsLruPerHash(CtableBase):
    def __init__(self, so, name, dTypes):
        super(CmapsLruPerHash, self).__init__(so, name, dTypes)


class structKsym(ct.Structure):
    _fields_ = [("addr", ct.c_ulong), ("name", ct.c_char_p)]


class CmapsStack(CtableBase):
    def __init__(self, so, name, dTypes):
        super(CmapsStack, self).__init__(so, name, dTypes)
        self._so.ksym_search.restype = ct.POINTER(structKsym)
        self._so.ksym_search.argtypes = [ct.c_ulong]

    def getStacks(self, stack_id, sLen=-1):
        arr = []
        stks = self.getKeyValue(stack_id)
        if sLen == -1:
            sLen = len(stks)
        for i in range(sLen):
            if stks[i]:
                p = self._so.ksym_search(stks[i])
                arr.append(p.contents.name.decode())
            else:
                break
        return arr


class CmapsEvent(CeventBase):
    def __init__(self, so, name, dTypes):
        super(CmapsEvent, self).__init__(so, name)
        self.__d = dTypes['vtype']
        self._cb = None
        self._lostcb = None
        self.__setup_localCalls()

    def __setup_localCalls(self):
        self._so.lbc_event_loop.restype = ct.c_int
        self._so.lbc_event_loop.argtypes = [ct.c_int, ct.c_int]

    def open_perf_buffer(self, cb, lost=None):
        self._cb = cb

    def perf_buffer_poll(self, timeout=-1):
        t = self

        def _callback(context, cpu, data, size):
            if t._cb:
                t._cb(cpu, data, size)

        def _lostcb(context, cpu, count):
            if t._lostcb:
                t._cb(cpu, count)
            else:
                print("cpu%d lost %d events" % (cpu, count))

        eventCallback = ct.CFUNCTYPE(None, ct.c_void_p, ct.c_int, ct.c_void_p, ct.c_ulong)
        lostCallback = ct.CFUNCTYPE(None, ct.c_void_p, ct.c_int, ct.c_ulonglong)
        _cb = eventCallback(_callback)
        _lost = lostCallback(_lostcb)
        self._so.lbc_set_event_cb.restype = ct.c_int
        self._so.lbc_set_event_cb.argtypes = [ct.c_int, eventCallback, lostCallback]
        self._so.lbc_set_event_cb(self._id, _cb, _lost)
        self._so.lbc_event_loop(self._id, timeout)

    def event(self, data):
        e = CtypeData(self.__d, data)
        return e


def paserMaps(so, name, dTypes):
    t = dTypes['type']
    tDict = {'event': CmapsEvent,
             'hash': CmapsHash,
             'array': CmapsArray,
             'lruHash': CmapsLruHash,
             'perHash': CmapsPerHash,
             'perArray': CmapsPerArray,
             'lruPerHash': CmapsLruPerHash,
             'stack': CmapsStack, }
    if t in tDict:
        return tDict['t'](so, name, dTypes)


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
# cython:language_level=3
"""
-------------------------------------------------
   File Name：     lbcBase
   Description :
   Author :       liaozhaoyan
   date：          2021/7/20
-------------------------------------------------
   Change Activity:
                   2021/7/20:
-------------------------------------------------
"""
__author__ = 'liaozhaoyan'

import sys
import os
import base64
import ctypes as ct
import _ctypes as _ct
import json
import hashlib
from pylcc.lbcMaps import CmapsEvent, CmapsHash, CmapsArray, \
    CmapsLruHash, CmapsPerHash, CmapsPerArray, CmapsLruPerHash, CmapsStack
from surftrace.execCmd import CexecCmd
from surftrace.surfException import InvalidArgsException, RootRequiredException, FileNotExistException, DbException
from surftrace.lbcClient import ClbcClient

LBC_COMPILE_PORT = 7655
SEG_UNIT = 4096


def segDecode(stream):
    line = b""
    l = len(stream)
    for i in range(0, l, 4 * SEG_UNIT):
        s = stream[i:i + 4 * SEG_UNIT]
        line += base64.b64decode(s)
    if l % (4 * SEG_UNIT):
        i = int(l / (4 * SEG_UNIT) * (4 * SEG_UNIT))
        line += base64.b64decode(stream[i:])
    return line


class ClbcBase(object):
    def __init__(self, bpf, bpf_str="",
                 server="pylcc.openanolis.cn",
                 arch="", ver="", env="",
                 workPath=None, logLevel=-1):

        if "LBC_SERVER" in os.environ:
            server = os.environ["LBC_SERVER"]
        if "LBC_LOGLEVEL" in os.environ:
            logLevel = int(os.environ["LBC_LOGLEVEL"])
        if workPath:
            self._wPath = workPath
        else:
            self._wPath = os.getcwd()
        super(ClbcBase, self).__init__()
        self.__need_del = False
        self._server = server
        self._c = CexecCmd()
        self._checkRoot()
        self._env = env
        self._logLevel = logLevel

        if ver == "":
            ver = self._c.cmd('uname -r')
        if arch == "":
            arch = self._c.cmd('uname -m')
        self._checkBtf(ver, arch)
        bpf_so = self._getSo(bpf, bpf_str, ver, arch)

        self._loadSo(bpf_so)
        self.maps = {}
        self._loadMaps()

    def __del__(self):
        if self.__need_del:
            self.__so.lbc_bpf_exit()

    def _checkBtf(self, ver, arch):
        if os.path.exists('/sys/kernel/btf/vmlinux'):
            return
        name = "/boot/vmlinux-%s" % ver
        if not os.path.exists(name):
            cli = ClbcClient(server=self._server, ver=ver, arch=arch)
            dRecv = cli.getBtf()
            if dRecv['btf'] is None:
                print("get btf failed, log is:\n%s" % dRecv['log'])
                raise InvalidArgsException("get btf failed.")
            print("get btf from remote success.")
            with open(name, 'wb') as f:
                f.write(segDecode(dRecv['btf']))

    @staticmethod
    def _closeSo(so):
        _ct.dlclose(so._handle)

    def _getSo(self, bpf, s, ver, arch):
        bpf_so = self._wPath + '/' + bpf + ".so"
        need = False
        if s == "":
            bpf_c = self._wPath + '/' + bpf + ".bpf.c"
            if self._checkCCompile(bpf_c, bpf_so, ver, arch):
                with open(bpf_c, 'r') as f:
                    s = f.read()
                need = True
        else:
            need = self._checkStrCompile(s, bpf_so, ver, arch)
        if need:
            self._compileSo(s, bpf_so, ver, arch)
        return bpf_so

    def _checkCCompile(self, bpf_c, bpf_so, ver, arch):
        cFlag = os.path.exists(bpf_c)
        oFlag = os.path.exists(bpf_so)
        if not (cFlag or oFlag):  # is not exist
            raise FileNotExistException("bpf.c or so is not in this dictionary.")
        elif not oFlag and cFlag:  # only bpf.c
            return True
        elif oFlag and not cFlag:  # only so, should check version
            if self._checkVer(bpf_so, ver, arch):
                raise FileNotExistException("bad bpf.so and not bpf.c")
            return False
        else:  # both bpf.c and bo, check hash and version
            with open(bpf_c, "r") as f:
                s = f.read()
            s += self._env
            if sys.version_info.major >= 3:
                cHash = hashlib.sha256(s.encode()).hexdigest()
            else:
                cHash = hashlib.sha256(s).hexdigest()
            if self._checkHash(bpf_so, cHash):
                return True
            return self._checkVer(bpf_so, ver, arch)

    def _checkStrCompile(self, s, bpf_so, ver, arch):
        oFlag = os.path.exists(bpf_so)
        if not oFlag:  # only string
            return True
        else:  # both bpf.c and bo, check hash and version
            s += self._env
            if sys.version_info.major >= 3:
                cHash = hashlib.sha256(s.encode()).hexdigest()
            else:
                cHash = hashlib.sha256(s).hexdigest()
            if self._checkHash(bpf_so, cHash):
                return True
            return self._checkVer(bpf_so, ver, arch)

    def _parseVer(self, ver):
        major, minor, _ = ver.split(".", 2)
        return major

    def _checkVer(self, bpf_so, ver, arch):
        """if should compile return ture, else return false"""
        try:
            so = ct.CDLL(bpf_so)
        except (OSError, FileNotFoundError):
            return True
        so.lbc_get_map_types.restype = ct.c_char_p
        so.lbc_get_map_types.argtypes = []
        s = so.lbc_get_map_types()
        soVer = json.loads(s)['kern_version']
        self._closeSo(so)

        soMajor = self._parseVer(soVer)
        hMajor = self._parseVer(ver)
        return (int(soMajor) > 3) ^ (int(hMajor) > 3)

    def _checkHash(self, bpf_so, cHash):
        """if should compile return ture, else return false"""
        try:
            so = ct.CDLL(bpf_so)
        except (OSError, FileNotFoundError):
            return True
        so.lbc_get_map_types.restype = ct.c_char_p
        so.lbc_get_map_types.argtypes = []
        s = so.lbc_get_map_types()
        soHash = json.loads(s)['hash']
        self._closeSo(so)
        return not cHash == soHash

    def _checkRoot(self):
        cmd = 'whoami'
        line = self._c.cmd(cmd).strip()
        if line != "root":
            raise RootRequiredException('this app need run as root')

    def _compileSo(self, s, bpf_so, ver, arch):
        cli = ClbcClient(server=self._server, ver=ver, arch=arch, port=LBC_COMPILE_PORT)
        dRecv = cli.getC(s, self._env)
        if dRecv is None:
            raise Exception("receive error")
        if dRecv['so'] is None:
            print("compile failed, log is:\n%s" % dRecv['clog'])
            raise InvalidArgsException("compile failed.")
        print("remote server compile success.")
        with open(bpf_so, 'wb') as f:
            f.write(segDecode(dRecv['so']))

    def _loadSo(self, bpf_so):
        self.__need_del = True
        self.__so = ct.CDLL(bpf_so)
        self.__so.lbc_bpf_init.restype = ct.c_int
        self.__so.lbc_bpf_init.argtypes = [ct.c_int]
        r = self.__so.lbc_bpf_init(self._logLevel)
        if r != 0:
            self.__need_del = False
            raise InvalidArgsException("so init failed")

    def _loadMaps(self):
        self.__so.lbc_get_map_types.restype = ct.c_char_p
        self.__so.lbc_get_map_types.argtypes = []
        s = self.__so.lbc_get_map_types()
        d = json.loads(s)['maps']
        tDict = {'event': CmapsEvent,
                 'hash': CmapsHash,
                 'array': CmapsArray,
                 'lruHash': CmapsLruHash,
                 'perHash': CmapsPerHash,
                 'perArray': CmapsPerArray,
                 'lruPerHash': CmapsLruPerHash,
                 'stack': CmapsStack, }
        for k in d.keys():
            t = d[k]['type']
            if t in tDict:
                self.maps[k] = tDict[t](self.__so, k, d[k])
            else:
                raise InvalidArgsException("bad type: %s, key: %s" % (t, k))

    def getMap(self, name, data, size):
        stream = ct.string_at(data, size)
        try:
            return self.maps[name].event(stream)
        except IndexError:
            return None


if __name__ == "__main__":
    pass